/***************************************************************************
                             conn-http.h
                             -----------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This library is free software; you can redistribute it and/or           *
 * modify it under the terms of the GNU Library General Public             *
 * License as published by the Free Software Foundation; either            *
 * version 2 of the License, or (at your option) any later version.        *
 *                                                                         *
 * This library is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * Library General Public License for more details.                        *
 *                                                                         *
 * You should have received a copy of the GNU Library General Public       *
 * License along with this library; if not, write to the                   *
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,            *
 * Boston, MA  02111-1307, USA.                                            *
 *                                                                         *
 ***************************************************************************/

#ifndef _gnet_conn_http_h_included_
#define _gnet_conn_http_h_included_

#include <gnet.h>

#define  GNET_CONN_HTTP_FLAG_SKIP_HEADER_CHECK  1

 /* other methods are currently unsupported */
typedef enum
{
 GNET_CONN_HTTP_METHOD_GET,
 GNET_CONN_HTTP_METHOD_POST
} GConnHttpMethod;

/* The entire GConnHttp struct is opaque and private */
typedef struct _GConnHttp GConnHttp;


/***************************************************************************
 *                                                                         *
 *   Error codes (see GConnHttpEventError below). These usually signal     *
 *    internal errors only. Errors by the HTTP server will be transmitted  *
 *    via the RESPONSE event.                                              *
 *                                                                         *
 ***************************************************************************/

typedef enum
{
 GNET_CONN_HTTP_ERROR_UNSPECIFIED,
 GNET_CONN_HTTP_ERROR_PROTOCOL_UNSUPPORTED,
} GConnHttpError;


/***************************************************************************
 *                                                                         *
 *   Event structures - work similar to GdkEvent                           *
 *                                                                         *
 ***************************************************************************/

typedef enum
{
 GNET_CONN_HTTP_RESOLVED,           /* resolved hostname            */
 GNET_CONN_HTTP_CONNECTED,          /* connected to host            */
 GNET_CONN_HTTP_RESPONSE,           /* got response (incl. headers) */
 GNET_CONN_HTTP_REDIRECT,           /* got redirect                 */
 GNET_CONN_HTTP_DATA_PARTIAL,       /* we got some data             */
 GNET_CONN_HTTP_DATA_COMPLETE,      /* we got all data              */
 GNET_CONN_HTTP_TIMEOUT,            /* the connection timed out     */
 GNET_CONN_HTTP_ERROR,              /* GConnHttp problem            */
} GConnHttpEventType;

typedef struct _GConnHttpEvent         GConnHttpEvent;
typedef struct _GConnHttpEventResolved GConnHttpEventResolved;
typedef struct _GConnHttpEventResponse GConnHttpEventResponse;
typedef struct _GConnHttpEventRedirect GConnHttpEventRedirect;
typedef struct _GConnHttpEventData     GConnHttpEventData; 
typedef struct _GConnHttpEventError    GConnHttpEventError; 

struct _GConnHttpEvent
{
 GConnHttpEventType   type;           /* type of event (see above)         */
 gsize                stsize;         /* size of underlying event struct   */
 gpointer             padding[4];     /* padding for future expansion      */
};

struct _GConnHttpEventResolved
{
 GConnHttpEvent       parent;         /* GConnHttpEvent structure          */
 GInetAddr           *ia;             /* GInetAddr of the hostname         */
 gpointer             padding[4];     /* padding for future expansion      */
};

struct _GConnHttpEventRedirect
{
 GConnHttpEvent       parent;         /* GConnHttpEvent structure          */
 
 guint                num_redirects;  /* number of redirects so far        */
 guint                max_redirects;  /* max. num. of redirects allowed    */
 gchar               *new_location;   /* redirect location if provided     */
 gboolean             auto_redirect;  /* FALSE if action is needed         */
 
 gpointer             padding[4];     /* padding for future expansion      */
};

struct _GConnHttpEventResponse
{
 GConnHttpEvent       parent;         /* GConnHttpEvent structure          */

 guint                response_code;  /* response code, e.g. 200, or 404   */
 gchar              **header_fields;  /* NULL-terminated array of strings  */
 gchar              **header_values;  /* NULL-terminated array of strings  */

 gpointer             padding[8];     /* padding for future expansion      */
};

struct _GConnHttpEventData
{
 GConnHttpEvent       parent;         /* GConnHttpEvent structure          */

 guint64              content_length; /* set if available, otherwise 0     */
 guint64              data_received;  /* bytes received so far             */
 GString             *buffer;         /* buffer, can be emptied/truncated  */

 gpointer             padding[8];     /* padding for future expansion      */
};

struct _GConnHttpEventError
{
 GConnHttpEvent       parent;         /* GConnHttpEvent structure          */
 
 GConnHttpError       code;           /* error code                        */
 gchar               *message;        /* message (use for debugging only)  */
 
 gpointer             padding[4];     /* padding for future expansion      */
};

/***************************************************************************
 *                                                                         *
 *   GConnHttp callback function prototype                                 *
 *                                                                         *
 ***************************************************************************/

typedef void   (*GConnHttpFunc) (GConnHttp      *conn, 
                                 GConnHttpEvent *event, 
                                 gpointer        user_data);

/***************************************************************************
 *                                                                         *
 *   GConnHttp API functions                                               *
 *                                                                         *
 ***************************************************************************/

GConnHttp       *gnet_conn_http_new                (void);

gboolean         gnet_conn_http_set_uri            (GConnHttp        *conn,
                                                    const gchar      *uri);

gboolean         gnet_conn_http_set_header         (GConnHttp        *conn,
                                                    const gchar      *field,
                                                    const gchar      *value,
                                                    guint             flags);

void             gnet_conn_http_set_max_redirects  (GConnHttp        *conn,
                                                    guint             num);

void             gnet_conn_http_set_timeout        (GConnHttp        *conn,        // FIXME: implement
                                                    guint             timeout);

gboolean         gnet_conn_http_set_user_agent     (GConnHttp        *conn,
                                                    const gchar      *agent);

gboolean         gnet_conn_http_set_method         (GConnHttp        *conn,
                                                    GConnHttpMethod   method,
                                                    const gchar      *post_data,
                                                    gsize             post_data_len);

void             gnet_conn_http_run_async          (GConnHttp        *conn,
                                                    GConnHttpFunc     func,
                                                    gpointer          user_data);

gboolean         gnet_conn_http_run                (GConnHttp        *conn,
                                                    GConnHttpFunc     func,
                                                    gpointer          user_data);

gboolean         gnet_conn_http_steal_buffer       (GConnHttp        *conn,
                                                    gchar           **buffer,
                                                    gsize            *buflen);

void             gnet_conn_http_cancel             (GConnHttp        *conn);

void             gnet_conn_http_delete             (GConnHttp        *conn);

#endif /* _gnet_conn_http_h_included_ */

